#include <math.h>
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#define	INCH	25.4	/* 1[inch] = 25.4[mm] */

/* ねじの種類リスト	*/
enum
{
	METRIC_EXTERNAL,		/* メートルおねじ */
	METRIC_INTERNAL,		/* メートルめねじ */
	UNIFIED_EXTERNAL,		/* ユニファイおねじ */
	UNIFIED_INTERNAL,		/* ユニファイめねじ */
	WHITWORTH_EXTERNAL,	/* ウィットおねじ */
	WHITWORTH_INTERNAL,	/* ウィットめねじ */
	N_THREAD_TYPE				/* ねじの種類の総数 */
};

/* ねじの種類 */
guint	thread_type = METRIC_EXTERNAL;

/* 目盛の種類のリスト */
enum
{
	RADIUS_SCALE,			/* 半径目盛 */
	DIAMETER_SCALE,		/* 直径目盛 */
	N_SCALE						/* 目盛の種類の数 */
};

/* 目盛の種類 */
guint scale_type = RADIUS_SCALE;

/* 切込みを計算するための係数リスト */
gdouble	keisu[N_THREAD_TYPE];	

/* 半径目盛、直径目盛用の換算用係数リスト */
gdouble	scale[N_SCALE];

/* ネジ山の角度リスト */
gdouble	kakudo[N_THREAD_TYPE];

/* window1が生成されたときに呼び出されるハンドラ */
void
on_window1_realize                     (GtkWidget       *widget,
                                        gpointer         user_data)
{
	/*	切込み用係数リストを初期化する */
	keisu[METRIC_EXTERNAL]		=	0.649;	/* メートルおねじ */
	keisu[METRIC_INTERNAL]		=	0.58;		/* メートルめねじ */
	keisu[UNIFIED_EXTERNAL]		=	0.613;	/* ユニファイおねじ */
	keisu[UNIFIED_INTERNAL]		=	0.58;		/* ユニファイめねじ */
	keisu[WHITWORTH_EXTERNAL]	=	0.64;		/* ウィットおねじ */
	keisu[WHITWORTH_INTERNAL]	=	0.569;	/* ウィットめねじ */
	
	/* ネジ山の角度リストを初期化する */
	kakudo[METRIC_EXTERNAL]		=	(60.0*M_PI)/180.0;
	kakudo[METRIC_INTERNAL]		=	(60.0*M_PI)/180.0;
	kakudo[UNIFIED_EXTERNAL]	=	(60.0*M_PI)/180.0;
	kakudo[UNIFIED_INTERNAL]	=	(60.0*M_PI)/180.0;
	kakudo[WHITWORTH_EXTERNAL]=	(55.0*M_PI)/180.0;
	kakudo[WHITWORTH_INTERNAL]=	(55.0*M_PI)/180.0;
	
	/* 目盛の換算用係数リストを初期化する */
	scale[RADIUS_SCALE]		= 1.0;
	scale[DIAMETER_SCALE]	=	2.0;
}

/* comboboxType(ネジタイプ選択用コンボボックス)が生成
	されたときに呼び出されるハンドラ	*/
void
on_comboboxType_realize                (GtkWidget       *widget,
                                        gpointer         user_data)
{
	/* デフォルトのアイテムをアクティブにする */
	gtk_combo_box_set_active(GTK_COMBO_BOX(widget),thread_type);
}

/* comboboxType(ネジタイプ選択用コンボボックス)で
	選択状態が変化したときに呼び出されるハンドラ	*/
void
on_comboboxType_changed                (GtkComboBox     *combobox,
                                        gpointer         user_data)
{
	thread_type = gtk_combo_box_get_active(combobox);
	
	if(thread_type==METRIC_EXTERNAL | thread_type==METRIC_INTERNAL)
	{
		gtk_label_set_text(
			GTK_LABEL(lookup_widget(GTK_WIDGET(combobox),"labelKei")),
			"呼び径[mm]");
		
		gtk_label_set_text(
			GTK_LABEL(lookup_widget(GTK_WIDGET(combobox),"labelPitch")),
			"ピッチ[mm]");
	}
	
	else
	{
		gtk_label_set_text(
			GTK_LABEL(lookup_widget(GTK_WIDGET(combobox),"labelKei")),
			"呼び径[inch]");
		
		gtk_label_set_text(
			GTK_LABEL(lookup_widget(GTK_WIDGET(combobox),"labelPitch")),
			"山数");
	}
		
}

/* comboboxMemori(目盛タイプ選択用コンボボックス)が生成
	されたときに呼び出されるハンドラ	*/
void
on_comboboxMemori_realize              (GtkWidget       *widget,
                                        gpointer         user_data)
{
	/* デフォルトのアイテムをアクティブにする */
	gtk_combo_box_set_active(GTK_COMBO_BOX(widget),scale_type);
}


/* comboboxMemori(目盛タイプ選択用コンボボックス)で
	選択状態が変化したときに呼び出されるハンドラ	*/
void
on_comboboxMemori_changed              (GtkComboBox     *combobox,
                                        gpointer         user_data)
{
	scale_type = gtk_combo_box_get_active(combobox);
}


/* 「計算する」ボタンが押されたときに
	呼び出されるハンドラ	*/
void
on_buttonKeisan_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
	gchar			*text;
	gdouble	kei,pitch,kirikomi1;	/* 呼び径、ピッチ、初回切込み	*/
	gint			kaisu;							/* ネジ切り回数 */

	/* 呼び径を取得 */
	text = gtk_entry_get_text(
		GTK_ENTRY(lookup_widget(GTK_WIDGET(button),"entryKei")));
	kei = getKei(text);
	
	/* ピッチを取得 */
	text = gtk_entry_get_text(
		GTK_ENTRY(lookup_widget(GTK_WIDGET(button),"entryPitch")));
	pitch = getPitch(text);
	
	/* 初回切込みを取得 */
	text = gtk_entry_get_text(
		GTK_ENTRY(lookup_widget(GTK_WIDGET(button),"entryKirikomi")));
	kirikomi1 = g_strtod(text,NULL);
	
	/* 回数を取得 */
	text = gtk_entry_get_text(
		GTK_ENTRY(lookup_widget(GTK_WIDGET(button),"entryKaisu")));
	kaisu = (gint)g_strtod(text,NULL);
	
	/* 計算結果を取得 */
	text = get_kekka(kei,pitch,kirikomi1,kaisu);
	
	/* 結果をTextViewに表示 */
	gtk_text_buffer_set_text(
		gtk_text_view_get_buffer(
			GTK_TEXT_VIEW(
				lookup_widget(GTK_WIDGET(button),"textviewKekka"))), _(text), -1);
}

/* 文字列から径を求める
	ユニファイ、ウィットねじの場合
	径が1/4や1/8などと分数で表される */
gdouble getKei(gchar *text)
{
	gdouble	kei=0.0,bunshi,bunbo;
	char **array = g_strsplit(text,"/",2);
	if(array[0]!=NULL)
		bunshi=g_strtod(array[0],NULL);
	
	if(array[1]!=NULL)
	{
		bunbo=g_strtod(array[1],NULL);
		kei=(bunshi/bunbo);
	}
	
	else
			kei=bunshi;
	
	/* メートルねじの場合はそのまま径を返す　*/
	if((thread_type==METRIC_EXTERNAL)
			| (thread_type==METRIC_INTERNAL))
		return	kei;
	
	/* 25.4を掛けてmm単位に換算してから径を返す */
	return	kei*INCH;
}

/* 文字列からピッチ[mm]を求める
	ユニファイネジやウィットネジの場合、
	ピッチの代わりに1inch=25.4mmあたりの
	山数が使われるので、ピッチへ換算する */
gdouble getPitch(gchar *text)
{
	gdouble	pitch;
	pitch=g_strtod(text,NULL);
	
	if((thread_type==METRIC_EXTERNAL)
		| (thread_type==METRIC_INTERNAL))
		return	pitch;
	
	return	25.4/pitch;
}
		
/* 計算結果を計算し、文字列として返す */
gchar *get_kekka(gdouble kei,					/* 呼び径 [mm]*/
									gdouble pitch,				/* ピッチ [mm]*/
									gdouble kirikomi1,		/* 初回切込み [mm]*/
									gint kaisu)						/* 切込み回数 */
{
	gdouble	kirikomi;				/* 切込み [mm] */
	gdouble	kirikomi_pre;		/* 前回の切込み [mm]*/
	gdouble	kirikomi_end;		/* 最終切込み [mm] */
	gdouble	okurikomi=0;		/* バイトの横方向への移動量 [mm] */
	gchar			*kekka;					/* 計算結果 */
	gchar			*line;
	gdouble	d;
	gdouble	H;
	gint			i=1;
	
	/* 最終的な切込みは、
		メートルおねじの場合：0.649×ピッチ
		メートルめねじの場合：0.58 ×ピッチ
		ユニファイおねじの場合：0.613×ピッチ
		ユニファイめねじの場合：0.58×ピッチ
		ウィットおねじの場合：0.64×ピッチ
		ウィットめねじの場合：0.569×ピッチ
		となる。直径目盛の旋盤ではこの値を倍にする */
	kirikomi_end = keisu[thread_type]*scale[scale_type]*pitch;
	
	/* 最初の切り込み */
	kirikomi=kirikomi1;
	kekka=g_strdup_printf("%d回目:\t切込:%.2f\n",i,kirikomi);
	kirikomi_pre = kirikomi;
	
	/* 2回目以降の切り込みは、初回切込みと最終切込みの間を
		補間用の曲線で補間した値として求める */
	for(i=2; i<=kaisu; i++)
	{
		kirikomi=get_kirikomi(i,kaisu,kirikomi1,kirikomi_end);
		d=(kirikomi-kirikomi_pre)*tan(kakudo[thread_type]/2.0);
		if(d>=0.02)
		{
			okurikomi += d;
			line=g_strdup_printf("%d回目:\t切込:%.2f\t刃物送り:%.2f\n",i,kirikomi,okurikomi);
		}
		
		else
			line=g_strdup_printf("%d回目:\t切込:%.2f\n",i,kirikomi);
		
		kekka=g_strdup_printf("%s%s",kekka,line);
		kirikomi_pre = kirikomi;
	}
	
	/* めねじの場合は下穴径も計算する */
	switch(thread_type)
	{
		case	METRIC_EXTERNAL:		return	kekka;
		case	UNIFIED_EXTERNAL:		return	kekka;
		case	WHITWORTH_EXTERNAL:	return	kekka;
	}
	
	/* [ネジの下穴径]=[ねじの呼び径]-(5/4)*H
		ここで、
		H=P/{2*tan(θ/2)}
		P:ピッチ	θ:ねじ山の角度 */
	H=pitch/(2.0*tan(kakudo[thread_type]/2.0));
	line=g_strdup_printf("ねじの下穴径:\t%.2f\n",kei-(5.0*H)/4.0);
	kekka=g_strdup_printf("%s%s",line,kekka);
	
	return	kekka;
}

/* i番目の切り込みを求める */
gdouble get_kirikomi(gint i,
											gint kaisu,							/* 切込み回数 */
											gdouble kirikomi_1st,	/* 初回切込み */
											gdouble kirikomi_end)	/* 最終切込み */
{
	return g((gdouble)i,1.0,(gdouble)kaisu,kirikomi_1st,kirikomi_end);
}

/* (x1,g1)と(x2,g2)の２点を補間する曲線		*/
gdouble g(gdouble x,gdouble x1,gdouble x2,gdouble g1,gdouble g2)
{
	static gdouble a=0.28,b=100.0;
	return	g1+(g2-g1)*((f(x,a,b)-f(x1,a,b))/(f(x2,a,b)-f(x1,a,b)));
}

gdouble f(gdouble x,gdouble a,gdouble b)
{
	return	(1+exp(a*(x+b)))/(1-exp(a*(x+b)));
}
